Search.setIndex({docnames:["index"],envversion:53,filenames:["index.rst"],objects:{},objnames:{},objtypes:{},terms:{index:0,modul:0,page:0,search:0},titles:["Welcome to forallpeople\u2019s documentation!"],titleterms:{document:0,forallpeopl:0,indic:0,tabl:0,welcom:0}})